#include <cassert>
#include <exception>
#include <iostream>
#include <tuple>

#include "common.hh"
#include "bytes.hh"

#include <args.hxx>


using namespace blky;

int main(int argc, char** argv) {
  args::ArgumentParser parser(
      "liblocky command line tool",
      "liblocky allow you to embed bits into video data secretly");

  args::MapFlag<std::string, DataFlow> from {
      parser, kDataFlowList, "input layer specifier", {"from"}, kDataFlowMap, args::Options::Required};

  args::MapFlag<std::string, DataFlow> to {
      parser, kDataFlowList, "output layer specifier", {"to"}, kDataFlowMap, args::Options::Required};

  args::Group src_group {
    parser, "source specifier", args::Group::Validators::Xor, args::Options::Required};
  args::Flag src_stdin {src_group, "src-stdin", "read from stdin", {"src-stdin", "stdin"}};

  args::Group dst_group  {
    parser, "destination specifier", args::Group::Validators::Xor, args::Options::Required};
  args::Flag dst_stdout {dst_group, "dst-stdout", "write to stdout", {"dst-stdout", "stdout"}};

  args::Group param_group {
    parser, "processing parameters", args::Group::Validators::DontCare
  };
  args::ValueFlag<std::tuple<uint32_t, uint32_t>> param_feat_num {
    param_group,
    "int>0,int>0",
    "number of features",
    {"feature-num"},
    {16, 16}
  };
  args::ValueFlag<uint8_t> param_feat_bits {
    param_group,
    "int>0",
    "number of bits that can be represented by a single feature",
    {"feature-bits"},
    1
  };
  args::ValueFlag<uint32_t> param_feat_first {
    param_group,
    "int>=0",
    "an index of first feature. used when encoding",
    {"feature-first-index"},
    0
  };
  args::ValueFlag<uint64_t> param_feat_seed {
    param_group,
    "int>0",
    "seed value for feature hopping",
    {"feature-hopping-seed"},
    1
  };

  try {
    parser.ParseCLI(argc, argv);

    std::vector<uint8_t> bytes;
    std::vector<Feature> features;

    // read input
    switch (args::get(from)) {
    case kBytes:
      if (src_stdin) {
        std::string temp;
        std::cin >> temp;
        bytes = {temp.begin(), temp.end()};
      } else {
        throw std::runtime_error {"invalid source format for bytes"};
      }
      break;

    case kFeatures:
      if (src_stdin) {
        while (!std::cin.eof()) {
          features.push_back({});
          std::cin >> features.back();
        }
      } else {
        throw std::runtime_error {"invalid source format for bytes"};
      }
      break;

    case kVideo:
      assert(false);
    }

    if (args::get(from) < args::get(to)) {
      // execute encoding
      switch (args::get(from)) {
      case kBytes:
        if (args::get(to) == kBytes) break;
        features = BytesEncoder(
            bytes,
            args::get(param_feat_num),
            args::get(param_feat_bits),
            args::get(param_feat_first),
            args::get(param_feat_seed));
        /* fallthrough */

      case kFeatures:
        if (args::get(to) == kFeatures) break;
        // TODO embed into video
        assert(false);

      case kVideo:
        if (args::get(to) == kVideo) break;
        assert(false);
      }

    } else if (args::get(from) > args::get(to)) {
      // execute decoding
      switch (args::get(from)) {
      case kVideo:
        if (args::get(to) == kVideo) break;
        // TODO extract features // features = XX
        assert(false);

      case kFeatures:
        if (args::get(to) == kFeatures) break;
        // TODO bytes = DataDecoder(features);
        assert(false);
        /* fallthrough */

      case kBytes:
        if (args::get(to) == kBytes) break;
        assert(false);
      }
    }

    // output
    switch (args::get(to)) {
    case kBytes:
      if (dst_stdout) {
        std::cout << std::string {bytes.begin(), bytes.end()} << std::endl;
      } else {
        throw std::runtime_error {"invalid destination format for bytes"};
      }
      break;
    case kFeatures:
      if (dst_stdout) {
        for (auto& f : features) std::cout << f << "\n";
      } else {
        throw std::runtime_error {"invalid destination format for features"};
      }
      break;
    case kVideo:
      break;
    }

  } catch (const args::Help&) {
    std::cout << parser << std::endl;
    return 0;

  } catch (const args::ParseError& e) {
    std::cerr << e.what() << std::endl;
    std::cerr << parser   << std::endl;
    return 1;

  } catch (const args::ValidationError& e) {
    std::cerr << e.what() << std::endl;
    std::cerr << parser   << std::endl;
    return 1;

  } catch (const std::runtime_error& e) {
    std::cerr << "runtime error: " << e.what() << std::endl;
    return 1;
  }
  return 0;
}
